------------------------------------------------------
-- Author : kurapica.igas@gmail.com
-- Create Date 	: 2012/06/28
-- ChangeLog

------------------------------------------------------
local version = 4

if not IGAS:NewAddon("IGAS.Widget.Unit", version) then
	return
end

------------------------------------------------------
-- Header for System.Widget.Unit.*
------------------------------------------------------
import "System"
import "System.Widget"

namespace "System.Widget.Unit"

print = print
geterrorhandler = geterrorhandler

errorhandler = errorhandler or function(err)
	return geterrorhandler and geterrorhandler()(err) or print(err)
end

class "UnitList"
	extend "IFIterator"

	local rawget = rawget
	local rawset = rawset
	local pairs = pairs

	------------------------------------------------------
	-- Event Manager
	------------------------------------------------------
	_UnitListEventDistribution = _UnitListEventDistribution or {}
	_UnitListEventManager = _UnitListEventManager or CreateFrame("Frame")
	_UnitListEventManager:Hide()
	_UnitListEventManager:SetScript("OnEvent", function(self, event, ...)
		if _UnitListEventDistribution[event] then
			for lst in pairs(_UnitListEventDistribution[event]) do
				lst:ParseEvent(event, ...)
			end
		end
	end)

	------------------------------------------------------
	-- Method
	------------------------------------------------------
	script "OnUnitListChanged"

	------------------------------------------------------
	-- Method
	------------------------------------------------------
	------------------------------------
	--- Parse event, Overridable
	-- @name ParseEvent
	-- @class function
	-- @param event
	-- @param unit
	-- @param ... args
	------------------------------------
	function ParseEvent(self, event, unit, ...)
		return self:EachK(unit, "Refresh", ...)
	end

	------------------------------------
	--- Register a event
	-- @name RegisterEvent
	-- @class function
	-- @param event the event's name
	-- @return nil
	-- @Usage IFModule:RegisterEvent("CUSTOM_EVENT_1")
	------------------------------------
	function RegisterEvent(self, event)
		_UnitListEventManager:RegisterEvent(event)
		_UnitListEventDistribution[event] = _UnitListEventDistribution[event] or {}
		_UnitListEventDistribution[event][self] = true
	end

	------------------------------------
	--- Check if an event is registered
	-- @name IsEventRegistered
	-- @class function
	-- @param event the event's name
	-- @return true if the event is registered to that Module
	------------------------------------
	function IsEventRegistered(self, event)
		return _UnitListEventDistribution[event] and _UnitListEventDistribution[event][self]
	end

	------------------------------------
	--- Undo register all events
	-- @name UnregisterAllEvents
	-- @class function
	------------------------------------
	function UnregisterAllEvents(self)
		for dis in pairs(_UnitListEventDistribution) do
			dis[self] = nil
		end
	end

	------------------------------------
	--- Undo Register a event
	-- @name UnregisterEvent
	-- @class function
	-- @param event the event's name
	------------------------------------
	function UnregisterEvent(self, event)
		if _UnitListEventDistribution[event] then
			_UnitListEventDistribution[event][self] = nil
		end
	end

	------------------------------------
	--- Get the next element, Overridable
	-- @name Next
	-- @class function
	-- @param key
	-- @return nextFunc
	-- @return self
	-- @return firstKey
	------------------------------------
	function Next(self, key)
		return _UnitList_Traverse[self], self, type(key) == "string" and key:lower() or nil
	end

	_UnitList_Info = _UnitList_Info or {}
	_UnitList_Prev = _UnitList_Prev or {}
	_UnitList_Next = _UnitList_Next or {}
	_UnitList_Traverse = _UnitList_Traverse or {}

	------------------------------------------------------
	-- Constructor
	------------------------------------------------------
	function UnitList(name)
		if type(name) ~= "string" then
			error("Usage : UnitList(name) - name must be a unique string.", 2)
		end

		local lst = {}
		local prev = "__UnitList_" .. name .. "_Prev"
		local nxt = "__UnitList_" .. name .. "_Next"

		_UnitList_Info[name] = lst
		_UnitList_Prev[lst] = prev
		_UnitList_Next[lst] = nxt
		_UnitList_Traverse[lst] = function (data, key)
			if type(key) ~= "table" then
				return data[key], data[key]
			else
				return key[nxt], key[nxt]
			end
		end

		return lst
	end

	------------------------------------------------------
	-- __exist
	------------------------------------------------------
	function __exist(name)
		if type(name) == "string" and _UnitList_Info[name] then
			return _UnitList_Info[name]
		end
	end

	------------------------------------------------------
	-- MetaMethod
	------------------------------------------------------
	function __index(self, unit)
		if type(unit) == "string" then
			return rawget(self, unit:lower())
		end
	end

	function __newindex(self, frame, unit)
		if type(frame) == "string" and type(unit) == "function" then
			rawset(self, frame, unit)
			return
		end

		if type(frame) ~= "table" then
			error("key must be a table.", 2)
		end

		if unit ~= nil and ( type(unit) ~= "string" or unit:match("^__") ) then
			error("value not supported.", 2)
		end

		unit = unit and unit:lower()

		if UnitList[unit] then
			error("value not supported.", 2)
		end

		local preKey = _UnitList_Prev[self]
		local nxtKey = _UnitList_Next[self]

		local prev = frame[preKey]
		local next = frame[nxtKey]
		local header = prev

		while type(header) == "table" do
			header = header[preKey]
		end

		-- no change
		if header == unit then
			return
		end

		-- Remove link
		if header then
			if prev == header then
				rawset(self, header, next)
				if next then next[preKey] = prev end
			else
				prev[nxtKey] = next
				if next then next[preKey] = prev end
			end

			frame[preKey] = nil
			frame[nxtKey] = nil
		end

		-- Add link
		if unit then
			if self[unit] then
				local tail = self[unit]

				while tail[nxtKey] do
					tail = tail[nxtKey]
				end

				tail[nxtKey] = frame
				frame[preKey] = tail
			else
				rawset(self, unit, frame)
				frame[preKey] = unit
			end
		end

		self:OnUnitListChanged()
	end

	function __call(self, unit)
		return _UnitList_Traverse[self], self, type(unit) == "string" and unit:lower() or nil
	end
endclass "UnitList"